from typing import Any, Tuple, Dict

from . import app as app, canvas as canvas, clock as clock, customtypes as customtypes, font as font, gl as gl, \
     graphics as graphics, gui as gui, image as image, input as input, math as math, media as media, model as model, \
     resource as resource, shapes as shapes, sprite as sprite, text as text, window as window

version: str
MIN_PYTHON_VERSION: Tuple[int, int]
MIN_PYTHON_VERSION_STR: str
compat_platform: str
env: str
value: str
options: Dict[str, Any]
